% Posterior mode estimator
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%   Thermodynamically Consistent Bayesian Analysis    %%%%%%%%%%%
%%%%%%%%%%%       of Closed Biochemical Reaction Systems        %%%%%%%%%%%
%%%%%%%%%%%            Posterior Mode Rate Estimation           %%%%%%%%%%%
%%%%%%%%%%%         Authors: Jenkinson, Zhong, Goutsias         %%%%%%%%%%%
%%%%%%%%%%%              Last Modified: 11/04/2010              %%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%

function [mode,obj,ER] = pmode(startingKappa,M,N,I,stoich,logkappa_0,tau,...
                            lambda,a,b,y,yp,x0,x0p,pert,tq,tol_low,...
                            tol_high,svdtol,timings,W,permMatM)
                        
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%% Initialization
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%

Q = size(y,1)-1;             % number of time points
P = sum(pert(1:N));          % number of perturbations

%
% Calculate singular value decomposition of matrix Xi
%

Xi      = stoich'*stoich;
[U,S,U] = svd(Xi); 
S0      = diag(S(S(1:M,1:M) > svdtol*max(max(S))));
M0      = M - size(S0,1);
U0      = U(1:M,1:M-M0);
Xi      = U0*diag(1./diag(S0))*U0';

%
% Calculate ytilde
%

yst      = repmat(0,[P+1,N]);
yst(1,:) = y(Q+1,:);             % steady-state data from reference system

if P > 0
    for p = 1:P
        yst(p+1,:) = yp(P*Q+p,:);    % steady-state data from perturbed systems
    end
end
ytilde = sum(yst,1)*stoich./(P+1);
ytilde = ytilde';

%
% Feasible initial rate constant values
%

[kappaFeas,kappaFree] = imposeWcond(startingKappa,W,permMatM);
kappaFreeStarting     = kappaFree;

%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%% SPSA Initialization
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%

A        = I/10;
delta    = tol_low;
s0       = 0.01;
magkappa = norm(kappaFree,1);
L0       = 10;
maxObj   = -inf;

%
% Calculate initial average gradient magnitude for gamma
%

gmag = 0;

for l = 1:L0
    epsilon     = 2.*round(rand(size(kappaFree)))-1;
    kappa_plus  = kappaFree + delta*epsilon;
    kappa_minus = kappaFree - delta*epsilon;
    C_plus  = object(y,yp,x0,x0p,pert,kappa_plus,N,Xi,ytilde,logkappa_0,...
                     lambda,tau,a,b,[tol_low;tol_low],timings,tq,W,permMatM);
    C_minus = object(y,yp,x0,x0p,pert,kappa_minus,N,Xi,ytilde,logkappa_0,...
                     lambda,tau,a,b,[tol_low;tol_low],timings,tq,W,permMatM);
    gmag = norm((C_plus-C_minus)./(2*delta*epsilon),1) + gmag;
end

gamma = L0*s0*magkappa*(1+A)^0.602/gmag;

%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% SPSA iterations
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                

wtbar = waitbar(0,sprintf('SPSA calculations. Please wait...'));

for iter = 0:(I-1)
    
    try
        waitbar(iter/I,wtbar);

        %
        % Estimate SPSA gradient
        %

        gamma_i     = gamma/(iter+1+A)^(0.602);
        delta_i     = delta/(iter+1)^(0.101);
        epsilon     = 2*round(rand(size(kappaFree)))-1;
        kappa_plus  = kappaFree + delta_i*epsilon;
        kappa_minus = kappaFree - delta_i*epsilon;

        C_plus  = object(y,yp,x0,x0p,pert,kappa_plus,N,Xi,ytilde,logkappa_0,...
                         lambda,tau,a,b,[tol_low;tol_low],timings,tq,W,permMatM);
        C_minus = object(y,yp,x0,x0p,pert,kappa_minus,N,Xi,ytilde,logkappa_0,...
                         lambda,tau,a,b,[tol_low;tol_low],timings,tq,W,permMatM);

        grad = (C_plus - C_minus)./(2*delta_i*epsilon);

        %
        % Keep track of best kappa value and save objective calculation
        %

        obj = (C_plus + C_minus)/2;
        if obj > maxObj
            maxObj   = obj;
            maxKappa = kappaFree;
        end

        %
        % Update kappa
        %

        kappaFree = kappaFree + gamma_i*grad;

        if sum(isnan(kappaFree)) > 0
            fprintf('\nNumerical problem in SPSA -- Returning to starting point\n');
            kappaFree = kappaFreeStarting;
        end

    catch ER % if there is an error, return to the starting point
        fprintf('\nNumerical problem in SPSA -- Returning to starting point\n');
        kappaFree = kappaFreeStarting;
    end
    
end

close(wtbar);

%
% Evaluate final objective function value and see if it is the max obtained 
%

obj = object(y,yp,x0,x0p,pert,kappaFree,N,Xi,ytilde,logkappa_0,lambda,tau,...
             a,b,[tol_low;tol_low],timings,tq,W,permMatM);
                     
if obj > maxObj
    maxObj   = obj;
    maxKappa = kappaFree;
end

%
% return final values
%

[mode,maxKappa] = imposeWcond(maxKappa,W,permMatM);
obj  = maxObj;